@@ -203,6 +203,15 @@ impl JumpRope {
     /// }
     /// assert_eq!(string, "Greetings!");
     /// ```
+    ///
+    /// Or more simply:
+    ///
+    /// ```
+    /// # use jumprope::*;
+    /// let rope = JumpRope::from("xxxGreetings!xxx");
+    /// let string = rope.slice_chunks(3..13).map(|(str, _len)| str).collect::<String>();
+    /// assert_eq!(string, "Greetings!");
+    /// ```
     pub fn slice_chunks(&self, range: Range<usize>) -> ContentRangeIter {
         let cursor = self.cursor_at_char(range.start, false);
         let node = unsafe { cursor.here_ptr().as_ref().unwrap() };
@@ -244,9 +253,9 @@ impl JumpRope {
 
 #[cfg(test)]
 mod tests {
+    use crate::fast_str_tools::*;
     use crate::JumpRope;
     use crate::jumprope::NODE_STR_SIZE;
-    use crate::utils::{count_chars, str_chars_to_bytes};
 
     fn check(rope: &JumpRope) {
         for (s, len) in rope.chunks() {
@@ -268,7 +277,7 @@ mod tests {
             let iter = rope.slice_chars(start..rope.len_chars());
             let str = iter.collect::<String>();
 
-            let byte_start = str_chars_to_bytes(&s, start);
+            let byte_start = char_to_byte_idx(&s, start);
             assert_eq!(str, &s[byte_start..]);
         }
     }
