@@ -16,8 +16,9 @@ use std::fmt::{Debug, Display, Formatter};
 use std::ops::Range;
 use rand::prelude::*;
 use rand::Rng;
+use crate::fast_str_tools::*;
 use crate::gapbuffer::GapBuffer;
-use crate::utils::*;
+// use crate::utils::*;
 // use crate::params::*;
 
 // Must be <= UINT16_MAX. Benchmarking says this is pretty close to optimal
@@ -62,6 +63,7 @@ pub struct JumpRope {
 
     // The first node is inline. The height is the max height we've ever used in the rope + 1. The
     // highest entry points "past the end" of the list, including the entire list length.
+    // TODO: Get rid of this and just rely on nexts out of here.
     pub(super) head: Node,
 
     // This is so dirty. The first node is embedded in JumpRope; but we need to allocate enough room
@@ -100,6 +102,9 @@ pub(super) struct SkipEntry {
     /// The number of *characters* between the start of the current node and the start of the next
     /// node.
     pub(super) skip_chars: usize,
+
+    #[cfg(feature = "wchar_conversion")]
+    pub(super) skip_pairs: usize,
 }
 
 // Make sure nexts uses correct alignment. This should be guaranteed by repr(C)
@@ -121,7 +126,12 @@ fn random_height(rng: &mut RopeRng) -> u8 {
 
 impl SkipEntry {
     fn new() -> Self {
-        SkipEntry { node: ptr::null_mut(), skip_chars: 0 }
+        SkipEntry {
+            node: ptr::null_mut(),
+            skip_chars: 0,
+            #[cfg(feature = "wchar_conversion")]
+            skip_pairs: 0
+        }
     }
 }
 
@@ -152,6 +162,7 @@ impl Node {
 
     fn alloc_with_height(height: u8, content: &str) -> *mut Node {
         //println!("height {} {}", height, max_height());
+        #![allow(clippy::manual_range_contains)]
         assert!(height >= 1 && height <= MAX_HEIGHT_U8);
 
         unsafe {
@@ -197,12 +208,18 @@ impl Node {
     pub(super) fn num_chars(&self) -> usize {
         self.first_next().skip_chars
     }
+
+    #[cfg(feature = "wchar_conversion")]
+    pub(super) fn num_surrogate_pairs(&self) -> usize {
+        self.first_next().skip_pairs
+    }
 }
 
 #[derive(Debug, Clone)]
 pub(crate) struct RopeCursor([SkipEntry; MAX_HEIGHT+1]);
 
 impl RopeCursor {
+    #[cfg(not(feature = "wchar_conversion"))]
     fn update_offsets(&mut self, height: usize, by: isize) {
         for i in 0..height {
             unsafe {
@@ -215,10 +232,27 @@ impl RopeCursor {
             }
         }
     }
+    #[cfg(feature = "wchar_conversion")]
+    fn update_offsets_wchar(&mut self, height: usize, by_chars: isize, by_pairs: isize) {
+        for i in 0..height {
+            unsafe {
+                // This is weird but makes sense when you realise the nexts in
+                // the cursor are pointers into the elements that have the
+                // actual pointers.
+                // Also adding a usize + isize is awful in rust :/
+                let entry = &mut (*self.0[i].node).nexts_mut()[i];
+                entry.skip_chars = entry.skip_chars.wrapping_add(by_chars as usize);
+                entry.skip_pairs = entry.skip_pairs.wrapping_add(by_pairs as usize);
+            }
+        }
+    }
 
-    fn move_within_node(&mut self, height: usize, by: isize) {
+    fn move_within_node(&mut self, height: usize, by_chars: isize, #[cfg(feature = "wchar_conversion")] by_pairs: isize) {
         for e in &mut self.0[..height] {
-            e.skip_chars = e.skip_chars.wrapping_add(by as usize);
+            e.skip_chars = e.skip_chars.wrapping_add(by_chars as usize);
+            #[cfg(feature = "wchar_conversion")] {
+                e.skip_pairs = e.skip_pairs.wrapping_add(by_pairs as usize);
+            }
         }
     }
 
@@ -230,6 +264,12 @@ impl RopeCursor {
         self.0[head_height as usize - 1].skip_chars
     }
 
+    #[cfg(feature = "wchar_conversion")]
+    pub(crate) fn wchar_pos(&self, head_height: u8) -> usize {
+        let entry = &self.0[head_height as usize - 1];
+        entry.skip_chars + entry.skip_pairs
+    }
+
     pub(crate) fn local_char_pos(&self) -> usize {
         self.0[0].skip_chars
     }
@@ -315,6 +355,21 @@ impl JumpRope {
         self.head.nexts()[self.head.height as usize - 1].skip_chars
     }
 
+    /// String length in wide characters (as would be reported by javascript / C# / etc).
+    ///
+    /// The byte length of this string when encoded to UTF16 will be exactly
+    /// `rope.len_wchars() * 2`.
+    #[cfg(feature = "wchar_conversion")]
+    pub fn len_wchars(&self) -> usize {
+        let SkipEntry {
+            skip_chars,
+            skip_pairs: skip_surrogate_pairs,
+            ..
+        } = self.head.nexts()[self.head.height as usize - 1];
+
+        skip_surrogate_pairs + skip_chars
+    }
+
     // Internal function for navigating to a particular character offset in the rope.  The function
     // returns the list of nodes which point past the position, as well as offsets of how far into
     // their character lists the specified characters are.
@@ -326,6 +381,9 @@ impl JumpRope {
         
         let mut offset = char_pos; // How many more chars to skip
 
+        #[cfg(feature = "wchar_conversion")]
+        let mut surrogate_pairs = 0; // Current wchar pos from the start of the rope
+
         let mut iter = RopeCursor([SkipEntry::new(); MAX_HEIGHT+1]);
 
         loop { // while height >= 0
@@ -336,16 +394,85 @@ impl JumpRope {
                 // Go right.
                 assert!(e == &self.head || !en.str.is_empty());
                 offset -= skip;
+                #[cfg(feature = "wchar_conversion")] {
+                    surrogate_pairs += next.skip_pairs;
+                }
                 e = next.node;
                 assert!(!e.is_null(), "Internal constraint violation: Reached rope end prematurely");
             } else {
                 // Record this and go down.
                 iter.0[height] = SkipEntry {
-                    skip_chars: offset,
                     node: e as *mut Node, // This is pretty gross
+                    skip_chars: offset,
+                    #[cfg(feature = "wchar_conversion")]
+                    skip_pairs: surrogate_pairs
                 };
 
-                if height == 0 { break; } else { height -= 1; }
+                if height != 0 {
+                    height -= 1;
+                } else {
+                    #[cfg(feature = "wchar_conversion")] {
+                        // Add on the wchar length at the current node.
+                        surrogate_pairs += en.str.count_surrogate_pairs(offset);
+                        if surrogate_pairs > 0 {
+                            for entry in &mut iter.0[0..self.head.height as usize] {
+                                entry.skip_pairs = surrogate_pairs - entry.skip_pairs;
+                            }
+                        }
+                    }
+                    break;
+                }
+            }
+        };
+
+        assert!(offset <= NODE_STR_SIZE);
+        iter
+    }
+
+    /// Create a cursor pointing wchar characters into the rope
+    #[cfg(feature = "wchar_conversion")]
+    pub(crate) fn cursor_at_wchar(&self, wchar_pos: usize, stick_end: bool) -> RopeCursor {
+        assert!(wchar_pos <= self.len_wchars());
+
+        let mut e: *const Node = &self.head;
+        let mut height = self.head.height as usize - 1;
+
+        let mut offset = wchar_pos; // How many more chars to skip
+
+        let mut char_pos = 0; // Char pos from the start of the rope
+
+        let mut iter = RopeCursor([SkipEntry::new(); MAX_HEIGHT+1]);
+
+        loop {
+            let en = unsafe { &*e };
+            let next = en.nexts()[height];
+            let skip = next.skip_chars + next.skip_pairs;
+            if offset > skip || (!stick_end && offset == skip && !next.node.is_null()) {
+                // Go right.
+                assert!(e == &self.head || !en.str.is_empty());
+                offset -= skip;
+                char_pos += next.skip_chars;
+                e = next.node;
+                assert!(!e.is_null(), "Internal constraint violation: Reached rope end prematurely");
+            } else {
+                // Record this and go down.
+                iter.0[height] = SkipEntry {
+                    node: e as *mut Node, // This is pretty gross
+                    skip_chars: char_pos,
+                    skip_pairs: offset
+                };
+
+                if height != 0 {
+                    height -= 1;
+                } else {
+                    char_pos += en.str.count_chars_in_wchars(offset);
+                    for entry in &mut iter.0[0..self.head.height as usize] {
+                        let skip_chars = char_pos - entry.skip_chars;
+                        entry.skip_chars = skip_chars;
+                        entry.skip_pairs -= skip_chars;
+                    }
+                    break;
+                }
             }
         };
 
@@ -356,7 +483,9 @@ impl JumpRope {
     fn cursor_at_start(&self) -> RopeCursor {
         RopeCursor([SkipEntry {
             node: &self.head as *const _ as *mut _,
-            skip_chars: 0
+            skip_chars: 0,
+            #[cfg(feature = "wchar_conversion")]
+            skip_pairs: 0,
         }; MAX_HEIGHT+1])
     }
 
@@ -366,10 +495,13 @@ impl JumpRope {
 
     // Internal fn to create a new node at the specified iterator filled with the specified
     // content.
-    unsafe fn insert_node_at(&mut self, cursor: &mut RopeCursor, contents: &str, num_chars: usize, update_cursor: bool) {
+    unsafe fn insert_node_at(&mut self, cursor: &mut RopeCursor, contents: &str, num_chars: usize, update_cursor: bool, #[cfg(feature = "wchar_conversion")] num_pairs: usize) {
         // println!("Insert_node_at {} len {}", contents.len(), self.num_bytes);
         // assert!(contents.len() < NODE_STR_SIZE);
         debug_assert_eq!(count_chars(contents), num_chars);
+        #[cfg(feature = "wchar_conversion")] {
+            debug_assert_eq!(count_utf16_surrogates(contents), num_pairs);
+        }
         debug_assert!(num_chars <= NODE_STR_SIZE);
 
         // TODO: Pin this sucka.
@@ -400,17 +532,31 @@ impl JumpRope {
             prev_skip.node = new_node;
             prev_skip.skip_chars = cursor.0[i].skip_chars;
 
+            #[cfg(feature = "wchar_conversion")] {
+                nexts[i].skip_pairs = num_pairs + prev_skip.skip_pairs - cursor.0[i].skip_pairs;
+                prev_skip.skip_pairs = cursor.0[i].skip_pairs;
+            }
+
             // & move the iterator to the end of the newly inserted node.
             if update_cursor {
                 cursor.0[i].node = new_node;
                 cursor.0[i].skip_chars = num_chars;
+                #[cfg(feature = "wchar_conversion")] {
+                    cursor.0[i].skip_pairs = num_pairs;
+                }
             }
         }
 
         for i in new_height..head_height {
             (*cursor.0[i].node).nexts_mut()[i].skip_chars += num_chars;
+            #[cfg(feature = "wchar_conversion")] {
+                (*cursor.0[i].node).nexts_mut()[i].skip_pairs += num_pairs;
+            }
             if update_cursor {
                 cursor.0[i].skip_chars += num_chars;
+                #[cfg(feature = "wchar_conversion")] {
+                    cursor.0[i].skip_pairs += num_pairs;
+                }
             }
         }
 
@@ -424,41 +570,51 @@ impl JumpRope {
         // skip. Figure out how much that is in bytes.
         let mut offset_bytes: usize = 0;
         // The insertion offset into the destination node.
-        let offset: usize = cursor.0[0].skip_chars;
+        let offset_chars: usize = cursor.0[0].skip_chars;
         let mut e = cursor.here_ptr();
 
+        let head_height = self.head.height as usize;
+
         // We might be able to insert the new data into the current node, depending on
         // how big it is. We'll count the bytes, and also check that its valid utf8.
         let num_inserted_bytes = contents.len();
-        let num_inserted_chars = count_chars(contents);
+        let mut num_inserted_chars = count_chars(contents);
+        #[cfg(feature = "wchar_conversion")]
+        let mut num_inserted_pairs = if num_inserted_bytes != num_inserted_chars {
+            count_utf16_surrogates(contents)
+        } else { 0 };
 
-        // Adding this short curcuit makes the code about 2% faster for 1% more code
-        if (*e).str.gap_start_chars as usize == offset && (*e).str.gap_len as usize >= num_inserted_bytes {
+        // Adding this short circuit makes the code about 2% faster for 1% more code
+        if (*e).str.gap_start_chars as usize == offset_chars && (*e).str.gap_len as usize >= num_inserted_bytes {
             // Short circuit. If we can just insert all the content right here in the gap, do so.
             (*e).str.insert_in_gap(contents);
-            cursor.update_offsets(self.head.height as usize, num_inserted_chars as isize);
-            cursor.move_within_node(self.head.height as usize, num_inserted_chars as isize);
+
+            #[cfg(feature = "wchar_conversion")] {
+                cursor.update_offsets_wchar(head_height, num_inserted_chars as isize, num_inserted_pairs as isize);
+                cursor.move_within_node(head_height, num_inserted_chars as isize, num_inserted_pairs as isize);
+            }
+            #[cfg(not(feature = "wchar_conversion"))] {
+                cursor.update_offsets(head_height, num_inserted_chars as isize);
+                cursor.move_within_node(head_height, num_inserted_chars as isize);
+            }
+
             self.num_bytes += num_inserted_bytes;
             return;
         }
 
-        if offset > 0 {
-            assert!(offset <= (*e).nexts()[0].skip_chars);
+        if offset_chars > 0 {
+            // Changing this to debug_assert reduces performance by a few % for some reason.
+            assert!(offset_chars <= (*e).nexts()[0].skip_chars);
             // This could be faster, but its not a big deal.
-            // let s = (*e).as_str();
-            // offset_bytes = str_get_byte_offset(s, offset);
-            offset_bytes = (*e).str.count_bytes(offset);
-
-            // println!("Offset {} offset_bytes {} s {:?}", offset, offset_bytes, s);
-            // let v: Vec<(usize, char)> = s.char_indices().collect();
-            // println!("{:?}", v);
+            offset_bytes = (*e).str.count_bytes(offset_chars);
         }
 
         // Can we insert into the current node?
         let current_len_bytes = (*e).str.len_bytes();
         let mut insert_here = current_len_bytes + num_inserted_bytes <= NODE_STR_SIZE;
 
-        // Can we insert into the subsequent node? Check if we're inserting at the end...
+        // If we can't insert here, see if we can move the cursor forward and insert into the
+        // subsequent node.
         if !insert_here && offset_bytes == current_len_bytes {
             // We can insert into the subsequent node if:
             // - We can't insert into the current node
@@ -468,13 +624,15 @@ impl JumpRope {
             if let Some(next) = (*e).first_next_mut().node.as_mut() {
                 if next.str.len_bytes() + num_inserted_bytes <= NODE_STR_SIZE {
                     offset_bytes = 0;
-                    // TODO: Try this on:
-                    // for e in &mut cursor.0[..next.height as usize] {
-                    //     e.node = next;
 
-                    for i in 0..next.height {
-                        cursor.0[i as usize].node = next;
-                        cursor.0[i as usize].skip_chars = 0;
+                    // Could do this with slice::fill but this seems slightly faster.
+                    for e in &mut cursor.0[..next.height as usize] {
+                        *e = SkipEntry {
+                            node: next,
+                            skip_chars: 0,
+                            #[cfg(feature = "wchar_conversion")]
+                            skip_pairs: 0
+                        };
                     }
                     e = next;
 
@@ -484,36 +642,47 @@ impl JumpRope {
         }
 
         if insert_here {
-            // println!("insert_here {}", contents);
             // First move the current bytes later on in the string.
             let c = &mut (*e).str;
             c.try_insert(offset_bytes, contents).unwrap();
 
             self.num_bytes += num_inserted_bytes;
             // .... aaaand update all the offset amounts.
-            cursor.update_offsets(self.head.height as usize, num_inserted_chars as isize);
-            cursor.move_within_node(self.head.height as usize, num_inserted_chars as isize);
+
+            #[cfg(feature = "wchar_conversion")] {
+                cursor.update_offsets_wchar(head_height, num_inserted_chars as isize, num_inserted_pairs as isize);
+                cursor.move_within_node(head_height, num_inserted_chars as isize, num_inserted_pairs as isize);
+            }
+            #[cfg(not(feature = "wchar_conversion"))] {
+                cursor.update_offsets(head_height, num_inserted_chars as isize);
+                cursor.move_within_node(head_height, num_inserted_chars as isize);
+            }
         } else {
             // There isn't room. We'll need to add at least one new node to the rope.
 
             // If we're not at the end of the current node, we'll need to remove
             // the end of the current node's data and reinsert it later.
             (*e).str.move_gap(offset_bytes);
-            // let trailing_data = (*e).str.end_as_str();
 
             let num_end_bytes = (*e).str.len_bytes() - offset_bytes;
             let mut num_end_chars: usize = 0;
+            #[cfg(feature = "wchar_conversion")]
+            let mut num_end_pairs: usize = 0;
             let end_str = if num_end_bytes > 0 {
                 // We'll truncate the node, but leave the bytes themselves there (for later).
 
                 // It would also be correct (and slightly more space efficient) to pack some of the
                 // new string's characters into this node after trimming it.
-                // let end_str = &(*e).as_str()[offset_bytes..];
                 let end_str = (*e).str.take_rest();
-                // (*e).num_bytes = offset_bytes as u8;
-                num_end_chars = (*e).num_chars() - offset;
+                num_end_chars = (*e).num_chars() - offset_chars;
 
-                cursor.update_offsets(self.head.height as usize, -(num_end_chars as isize));
+                #[cfg(feature = "wchar_conversion")] {
+                    num_end_pairs = (*e).num_surrogate_pairs() - (*e).str.gap_start_surrogate_pairs as usize;
+                    debug_assert_eq!(num_end_pairs, count_utf16_surrogates(end_str));
+                    cursor.update_offsets_wchar(head_height, -(num_end_chars as isize), -(num_end_pairs as isize));
+                }
+                #[cfg(not(feature = "wchar_conversion"))]
+                cursor.update_offsets(head_height, -(num_end_chars as isize));
                 self.num_bytes -= num_end_bytes;
                 Some(end_str)
             } else {
@@ -527,32 +696,45 @@ impl JumpRope {
             // utf8 codepoint.
             // let mut str_offset: usize = 0;
             let mut remainder = contents;
-            while !remainder.is_empty() {
+            // while !remainder.is_empty() {
+            loop {
                 // println!(". {}", remainder);
                 // Find the first index after STR_SIZE bytes
-                let mut byte_pos = 0;
-                let mut char_pos = 0;
 
-                // Find a suitable cut point. We should take as many characters as we can fit in
-                // the node, without splitting any unicode codepoints.
-                for c in remainder.chars() {
-                    // TODO: This could definitely be more efficient.
-                    let cs = c.len_utf8();
-                    if cs + byte_pos > NODE_STR_SIZE { break }
-                    else {
-                        char_pos += 1;
-                        byte_pos += cs;
+                if remainder.len() <= NODE_STR_SIZE {
+                    self.insert_node_at(cursor, remainder, num_inserted_chars, true, #[cfg(feature = "wchar_conversion")] num_inserted_pairs);
+                    break;
+
+                } else {
+                    // Find a suitable cut point. We should take as many characters as we can fit in
+                    // the node, without splitting any unicode codepoints.
+                    let mut byte_pos = NODE_STR_SIZE;
+                    loop { // Slide back to a character boundary.
+                        let c = remainder.as_bytes()[byte_pos];
+                        if c & 0b1100_0000 != 0b1000_0000 {
+                            break;
+                        }
+                        byte_pos -= 1;
                     }
+
+                    let slice = &remainder.as_bytes()[..byte_pos];
+                    let char_pos = count_chars_in_bytes(slice);
+                    num_inserted_chars -= char_pos;
+                    #[cfg(feature = "wchar_conversion")]
+                    let pairs = count_utf16_surrogates_in_bytes(slice);
+                    #[cfg(feature = "wchar_conversion")] {
+                        num_inserted_pairs -= pairs;
+                    }
+
+                    let (next, rem) = remainder.split_at(byte_pos);
+                    assert!(!next.is_empty());
+                    self.insert_node_at(cursor, next, char_pos, true, #[cfg(feature = "wchar_conversion")] pairs);
+                    remainder = rem;
                 }
-                
-                let (next, rem) = remainder.split_at(byte_pos);
-                assert!(!next.is_empty());
-                self.insert_node_at(cursor, next, char_pos, true);
-                remainder = rem;
             }
 
             if let Some(end_str) = end_str {
-                self.insert_node_at(cursor, end_str, num_end_chars, false);
+                self.insert_node_at(cursor, end_str, num_end_chars, false, #[cfg(feature = "wchar_conversion")] num_end_pairs);
             }
         }
 
@@ -561,31 +743,39 @@ impl JumpRope {
 
     unsafe fn del_at_cursor(&mut self, cursor: &mut RopeCursor, mut length: usize) {
         if length == 0 { return; }
-        let mut offset = cursor.local_char_pos();
+        let mut offset_chars = cursor.local_char_pos();
         let mut node = cursor.here_ptr();
         while length > 0 {
             {
                 let s = (&*node).first_next();
-                if offset == s.skip_chars {
+                if offset_chars == s.skip_chars {
                     // End of current node. Skip to the start of the next one.
                     node = s.node;
-                    offset = 0;
+                    offset_chars = 0;
                 }
             }
 
             let num_chars = (&*node).num_chars();
-            let removed = std::cmp::min(length, num_chars - offset);
+            let removed = std::cmp::min(length, num_chars - offset_chars);
             assert!(removed > 0);
 
+            // TODO: Figure out a better way to calculate this.
+            #[cfg(feature = "wchar_conversion")]
+            let removed_pairs = (*node).str.count_surrogate_pairs(offset_chars + removed)
+                - (*node).str.count_surrogate_pairs(offset_chars);
+
             let height = (*node).height as usize;
             if removed < num_chars || std::ptr::eq(node, &self.head) {
                 // Just trim the node down.
                 let s = &mut (*node).str;
-                let removed_bytes = s.remove_chars(offset, removed);
+                let removed_bytes = s.remove_chars(offset_chars, removed);
                 self.num_bytes -= removed_bytes;
 
                 for s in (*node).nexts_mut() {
                     s.skip_chars -= removed;
+                    #[cfg(feature = "wchar_conversion")] {
+                        s.skip_pairs -= removed_pairs;
+                    }
                 }
             } else {
                 // Remove the node from the skip list. This works because the cursor must be
@@ -596,6 +786,9 @@ impl JumpRope {
                     let s = &mut (*cursor.0[i].node).nexts_mut()[i];
                     s.node = (*node).nexts_mut()[i].node;
                     s.skip_chars += (*node).nexts()[i].skip_chars - removed;
+                    #[cfg(feature = "wchar_conversion")] {
+                        s.skip_pairs += (*node).nexts()[i].skip_pairs - removed_pairs;
+                    }
                 }
 
                 self.num_bytes -= (*node).str.len_bytes();
@@ -607,6 +800,9 @@ impl JumpRope {
             for i in height..self.head.height as usize {
                 let s = &mut (*cursor.0[i].node).nexts_mut()[i];
                 s.skip_chars -= removed;
+                #[cfg(feature = "wchar_conversion")] {
+                    s.skip_pairs -= removed_pairs;
+                }
             }
 
             length -= removed;
@@ -778,6 +974,8 @@ impl JumpRope {
     ///
     /// If the position names a location past the end of the rope, it is truncated.
     pub fn insert(&mut self, mut pos: usize, contents: &str) {
+        // if cfg!(debug_assertions) { self.check(); }
+
         if contents.is_empty() { return; }
         pos = std::cmp::min(pos, self.len_chars());
 
@@ -802,6 +1000,8 @@ impl JumpRope {
     /// assert_eq!(rope.to_string(), "Whoa!");
     /// ```
     pub fn remove(&mut self, mut range: Range<usize>) {
+        // if cfg!(debug_assertions) { self.check(); }
+
         range.end = range.end.min(self.len_chars());
         if range.start >= range.end { return; }
 
@@ -868,6 +1068,9 @@ impl JumpRope {
         let skip_over = &self.nexts[self.head.height as usize - 1];
         // println!("Skip over skip chars {}, num bytes {}", skip_over.skip_chars, self.num_bytes);
         assert!(skip_over.skip_chars <= self.num_bytes as usize);
+        #[cfg(feature = "wchar_conversion")] {
+            assert!(skip_over.skip_pairs <= skip_over.skip_chars);
+        }
         assert!(skip_over.node.is_null());
 
         // The offsets store the total distance travelled since the start.
@@ -879,6 +1082,8 @@ impl JumpRope {
 
         let mut num_bytes: usize = 0;
         let mut num_chars = 0;
+        #[cfg(feature = "wchar_conversion")]
+        let mut num_pairs = 0;
 
         for n in self.node_iter() {
             // println!("visiting {:?}", n.as_str());
@@ -891,25 +1096,42 @@ impl JumpRope {
             for (i, entry) in iter[0..n.height as usize].iter_mut().enumerate() {
                 assert_eq!(entry.node as *const Node, n as *const Node);
                 assert_eq!(entry.skip_chars, num_chars);
+                #[cfg(feature = "wchar_conversion")] {
+                    assert_eq!(entry.skip_pairs, num_pairs);
+                }
 
                 // println!("replacing entry {:?} with {:?}", entry, n.nexts()[i].node);
                 entry.node = n.nexts()[i].node;
                 entry.skip_chars += n.nexts()[i].skip_chars;
+                #[cfg(feature = "wchar_conversion")] {
+                    entry.skip_pairs += n.nexts()[i].skip_pairs;
+                }
             }
 
             num_bytes += n.str.len_bytes();
             num_chars += n.num_chars();
+
+            #[cfg(feature = "wchar_conversion")] {
+                assert_eq!(n.num_surrogate_pairs(), n.str.count_surrogate_pairs(n.num_chars()));
+                num_pairs += n.num_surrogate_pairs();
+            }
         }
 
         for entry in iter[0..self.head.height as usize].iter() {
             // println!("{:?}", entry);
             assert!(entry.node.is_null());
             assert_eq!(entry.skip_chars, num_chars);
+            #[cfg(feature = "wchar_conversion")] {
+                assert_eq!(entry.skip_pairs, num_pairs);
+            }
         }
 
         // println!("self bytes: {}, count bytes {}", self.num_bytes, num_bytes);
         assert_eq!(self.num_bytes, num_bytes);
         assert_eq!(self.len_chars(), num_chars);
+        #[cfg(feature = "wchar_conversion")] {
+            assert_eq!(self.len_wchars(), num_chars + num_pairs);
+        }
     }
 
     /// This method counts the number of bytes of memory allocated in the rope. This is purely for
@@ -938,12 +1160,16 @@ impl JumpRope {
     }
 
     #[allow(unused)]
+    // pub fn print(&self) {
     pub(crate) fn print(&self) {
         println!("chars: {}\tbytes: {}\theight: {}", self.len_chars(), self.num_bytes, self.head.height);
 
         print!("HEAD:");
         for s in self.head.nexts() {
             print!(" |{} ", s.skip_chars);
+            #[cfg(feature = "wchar_conversion")] {
+                print!("({}) ", s.skip_pairs);
+            }
         }
         println!();
 
@@ -951,8 +1177,81 @@ impl JumpRope {
             print!("{}:", i);
             for s in node.nexts() {
                 print!(" |{} ", s.skip_chars);
+                #[cfg(feature = "wchar_conversion")] {
+                    print!("({}) ", s.skip_pairs);
+                }
             }
-            println!("      : {:?} + {:?}", node.as_str_1(), node.as_str_2());
+            println!("      : {:?}(s{}) + {:?}(s{})",
+                     node.as_str_1(), count_utf16_surrogates(node.as_str_1()),
+                     node.as_str_2(), count_utf16_surrogates(node.as_str_2())
+            );
         }
     }
 }
+
+#[cfg(feature = "wchar_conversion")]
+impl JumpRope {
+    /// Convert from a unicode character count to a wchar index, like what you'd use in Javascript,
+    /// Java or C#.
+    pub fn chars_to_wchars(&self, chars: usize) -> usize {
+        let cursor = self.cursor_at_char(chars, true);
+        cursor.wchar_pos(self.head.height)
+    }
+
+    /// Convert a wchar index back to a unicode character count
+    pub fn wchars_to_chars(&self, wchars: usize) -> usize {
+        let cursor = self.cursor_at_wchar(wchars, true);
+        cursor.global_char_pos(self.head.height)
+    }
+
+    /// Insert the given utf8 string into the rope at the specified wchar position.
+    /// This is compatible with NSString, Javascript, etc.
+    ///
+    /// Returns the insertion position in characters.
+    pub fn insert_at_wchar(&mut self, mut pos_wchar: usize, contents: &str) -> usize {
+        pos_wchar = pos_wchar.min(self.len_wchars());
+
+        let mut cursor = self.cursor_at_wchar(pos_wchar, true);
+        // dbg!(pos_wchar, &cursor.0[0..3]);
+        unsafe { self.insert_at_cursor(&mut cursor, contents); }
+
+        debug_assert_eq!(
+            cursor.wchar_pos(self.head.height),
+            pos_wchar + count_chars(contents) + count_utf16_surrogates(contents)
+        );
+
+        cursor.global_char_pos(self.head.height)
+    }
+
+    /// Remove items from the rope, specified by the passed range. The indexes are interpreted
+    /// as wchar offsets (like you'd get in javascript / C# / etc).
+    pub fn remove_at_wchar(&mut self, mut range: Range<usize>) {
+        range.end = range.end.min(self.len_wchars());
+        if range.is_empty() { return; }
+
+        // Rather than making some fancy custom remove function, I'm just going to convert the
+        // removed range into a char range and delete that.
+        let cursor_end = self.cursor_at_wchar(range.end, true);
+        let char_end = cursor_end.global_char_pos(self.head.height);
+        drop(cursor_end);
+
+        // We need to stick_end so we can delete entries.
+        let mut cursor = self.cursor_at_wchar(range.start, true);
+        let char_start = cursor.global_char_pos(self.head.height);
+
+        unsafe { self.del_at_cursor(&mut cursor, char_end - char_start); }
+
+        debug_assert_eq!(cursor.wchar_pos(self.head.height), range.start);
+    }
+
+    /// Replace the characters in the specified wchar range with content.
+    pub fn replace_at_wchar(&mut self, range: Range<usize>, content: &str) {
+        // This is not optimized.
+        if !range.is_empty() {
+            self.remove_at_wchar(range.clone());
+        }
+        if !content.is_empty() {
+            self.insert_at_wchar(range.start, content);
+        }
+    }
+}
\ No newline at end of file
